package sun.nio.fs;

import java.nio.file.InvalidPathException;

class WindowsPathParser
{
    private static final String reservedChars = "<>:\"|?*";
    
    private WindowsPathParser() {
    }
    
    static WindowsPathParser.Result parse(String s) {
        return parse(s, true);
    }
    
    static WindowsPathParser.Result parseNormalizedPath(String s) {
        return parse(s, false);
    }
    
    private static WindowsPathParser.Result parse(String s, boolean b) {
        String str = "";
        WindowsPathType windowsPathType = null;
        int length = s.length();
        int n = 0;
        if (length > 1) {
            int char1 = s.charAt(0);
            int char2 = s.charAt(1);
            int n2 = 2;
            if (isSlash((char)char1) && isSlash((char)char2)) {
                windowsPathType = WindowsPathType.UNC;
                int nextNonSlash = nextNonSlash(s, n2, length);
                int nextSlash = nextSlash(s, nextNonSlash, length);
                if (nextNonSlash == nextSlash) {
                    throw new InvalidPathException(s, "UNC path is missing hostname");
                }
                String substring = s.substring(nextNonSlash, nextSlash);
                int nextNonSlash2 = nextNonSlash(s, nextSlash, length);
                int nextSlash2 = nextSlash(s, nextNonSlash2, length);
                if (nextNonSlash2 == nextSlash2) {
                    throw new InvalidPathException(s, "UNC path is missing sharename");
                }
                str = "\\\\" + substring + "\\" + s.substring(nextNonSlash2, nextSlash2) + "\\";
                n = nextSlash2;
            }
            else if (isLetter((char)char1) && char2 == ':') {
                int char3;
                if (length > 2 && isSlash((char)(char3 = s.charAt(2)))) {
                    if (char3 == '\\') {
                        str = s.substring(0, 3);
                    }
                    else {
                        str = s.substring(0, 2) + 92;
                    }
                    n = 3;
                    windowsPathType = WindowsPathType.ABSOLUTE;
                }
                else {
                    str = s.substring(0, 2);
                    n = 2;
                    windowsPathType = WindowsPathType.DRIVE_RELATIVE;
                }
            }
        }
        if (n == 0) {
            if (length > 0 && isSlash(s.charAt(0))) {
                windowsPathType = WindowsPathType.DIRECTORY_RELATIVE;
                str = "\\";
            }
            else {
                windowsPathType = WindowsPathType.RELATIVE;
            }
        }
        if (b) {
            StringBuilder o = new StringBuilder(s.length());
            o.append(str);
            return new WindowsPathParser.Result(windowsPathType, str, normalize(o, s, n));
        }
        return new WindowsPathParser.Result(windowsPathType, str, s);
    }
    
    private static String normalize(StringBuilder sb, String input, int i) {
        int length = input.length();
        i = (i = nextNonSlash(input, i, length));
        int n = '\0';
        while (i < length) {
            int char1 = input.charAt(i);
            if (isSlash((char)char1)) {
                if (n == ' ') {
                    throw new InvalidPathException(input, "Trailing char <" + (char)n + ">", i - 1);
                }
                sb.append(input, i, i);
                i = nextNonSlash(input, i, length);
                if (i != length) {
                    sb.append('\\');
                }
            }
            else {
                if (isInvalidPathChar((char)char1)) {
                    throw new InvalidPathException(input, "Illegal char <" + (char)char1 + ">", i);
                }
                n = char1;
                ++i;
            }
        }
        if (i != i) {
            if (n == ' ') {
                throw new InvalidPathException(input, "Trailing char <" + (char)n + ">", i - 1);
            }
            sb.append(input, i, i);
        }
        return sb.toString();
    }
    
    private static final boolean isSlash(char c) {
        return c == '\\' || c == '/';
    }
    
    private static final int nextNonSlash(String s, int index, int n) {
        while (index < n && isSlash(s.charAt(index))) {
            ++index;
        }
        return index;
    }
    
    private static final int nextSlash(String input, int n, int n2) {
        int char1;
        while (n < n2 && !isSlash((char)(char1 = input.charAt(n)))) {
            if (isInvalidPathChar((char)char1)) {
                throw new InvalidPathException(input, "Illegal character [" + (char)char1 + "] in path", n);
            }
            ++n;
        }
        return n;
    }
    
    private static final boolean isLetter(char c) {
        return (c >= 'a' && c <= 'z') || (c >= 'A' && c <= 'Z');
    }
    
    private static final boolean isInvalidPathChar(char ch) {
        return ch < ' ' || "<>:\"|?*".indexOf((int)ch) != -1;
    }
    
    static class Result
    {
        private final WindowsPathType type;
        private final String root;
        private final String path;
        
        Result(WindowsPathType type, String root, String path) {
            this.type = type;
            this.root = root;
            this.path = path;
        }
        
        WindowsPathType type() {
            return this.type;
        }
        
        String root() {
            return this.root;
        }
        
        String path() {
            return this.path;
        }
    }
}
